﻿:Class Fsm
⍝ Finite State Machine
⍝ This class takes a list of possible inputs and a table of states movements
⍝ and produces an instance of a machine it represents.
    
⍝ For example, the game of NIM with 1, 2 and 3 sticks where the aim is to
⍝ grab the last stick and any number of sticks in a row can be grabbed at
⍝ a single time (not more than 1 row though),
⍝ can be represented by a machine like this:
⍝ NIM (fresh game of 1, 2 and 3 sticks, the initial state)
⍝ there are 6 possible inputs:
⍝ 1. grab 1 stick  from the 1st row
⍝ 2. grab 1 stick  from the 2nd row
⍝ 3. grab 2 sticks from the 2nd row
⍝ 4. grab 1 stick  from the 3rd row
⍝ 5. grab 2 sticks from the 3rd row
⍝ 6. grab 3 sticks from the 3rd row
    
⍝ There are 24 possible states (2×3×4 combinations of sticks on each row)
⍝ but only a subset lead to a winning (final) state (000). Those are
⍝ 123, 022, 011, 101, 110, 0 (final)
    
⍝ The inputs are: 11, 12, 22, 13, 23, 33
⍝ The states are: 123, 022, 011, 101, 110, 000
⍝ The movements are (action/new state):
⍝       11     12     22    13     23     33
⍝ 123  13/22 33/110 23/101 11/22 22/101 12/110
⍝  22    -   13/11  23/0   12/11 22/0     -
⍝ 110  12/0  11/0     -      -     -      -
⍝ 101  13/0    -      -    11/0    -      -
⍝  11    -   13/0     -    12/0    -      -
    
⍝ To construct a Finite State Machine to represent that game you could use here
⍝ A←5 6⍴22 110 101 22 101 110,¯1 11 0 11 0 ¯1,0 0 ¯1 ¯1 ¯1 ¯1,0 ¯1 ¯1 0 ¯1 ¯1,¯1 0 ¯1 0 ¯1 ¯1
⍝ ⎕new Fsm ( (11 12 22 13 23 33) (123 22 11 101 110) A )
    
⍝ If output is desired we can use 
⍝ t←5 6⍴13 33 23 11 22 12,0 13 23 12 22 0,12 11 0 0 0 0,13 0 0 11 0 0,0 13 0 12 0 0
⍝ ⎕new Fsm ( (11 12 22 13 23 33) (123 22 11 101 110) A ¯1 t)
    
    ∇ init arg
      :Access public
      :Implements constructor
      ⎕SIGNAL(4 5 3∊⍨⍴,arg)↓5
      (Inputs States Movements InvalidState Actions)←5↑arg,(¯3+⍴arg)↓¯1,⊂''
     
      'Inputs not unique'⎕SIGNAL 11↓⍨Inputs≡∪Inputs
      'States not unique'⎕SIGNAL 11↓⍨States≡∪States
      'Illegal transition table'⎕SIGNAL 11↓⍨∧/,Movements∊States,0,InvalidState
      Movements,←InvalidState ⍝ cover bad input
      CurrentState←⍬⍴States ⋄ PreviousState←InvalidState
     
      ⎕DF'FSM ',⍕(⍴Inputs),{s/⍨×s←1↑⍵}⍴Actions
    ∇
    
    ∇ transit←{showoutput}Input actions;act;st;in;new
      :Access public
      :If 0=⎕NC'showoutput' ⋄ showoutput←~0∊⍴Actions ⋄ :EndIf
      transit←⍬
      :For act :In actions
          PreviousState←CurrentState
          transit,←new←Movements[st←States⍳CurrentState;in←Inputs⍳act]
         ⍝ Reset the machine on invalid or final input
          CurrentState←{⍵∊InvalidState:1↑States ⋄ 0=⍵:⍬⍴States ⋄ ⍵}new
          :If showoutput ⋄ (Actions,⊂'*Invalid/Reset*')[st;in],'→',new ⋄ :EndIf
      :EndFor
    ∇
    
    ∇ Example;A;O;s;i;t
      :Access public shared
      A←6 6⍴22 110 101 22 101 110,¯1 11 0 11 0 ¯1,0 0 ¯1 ¯1 ¯1 ¯1,0 ¯1 ¯1 0 ¯1 ¯1,¯1 0 ¯1 0 ¯1 ¯1,6⍴123
      t←6 6⍴13 33 23 11 22 12,0 13 23 12 22 0,12 11 0 0 0 0,13 0 0 11 0 0,0 13 0 12 0 0,6⍴¯1
      O←{⍕'I take'⍺'from'⍵}⌿0 10⊤t
      #.F←⎕NEW Fsm((i←11 12 22 13 23 33)(s←123 22 11 101 110 0)A ¯1 O)
      'Table is:' ⋄ (' ',s),i⍪t{⍺=0:'?' ⋄ (⍕⍺),'→',⍕⍵}¨A
      'Now try F.Input n-row(s) (e.g. 11 12); the moves wil be shown and the new states returned'
    ∇
    
:EndClass