﻿:Class Parser
⍝ Class to parse a string and set switches values accordingly
⍝ See Vector article in vol 19.4 for rules and logic behind.
    
⍝ This class has a few more features:
⍝ - the lowest generated error can be specified
⍝ - no space may be allowed before each switch (a la DOS)
⍝ - switch names can be UPPER or mixcased
⍝ - use " instead of '' (can be escaped with \")
⍝ To use these features call the class with a second string containing
⍝ 0 or more of 'error=nnn allownospace upper'
    
⍝ Example, create a parser accepting 3 UPPER switches: DATE which takes a value, SW2 which does
⍝ not take a value and SW3 which MAY take one of 'a', 'bc' or 'def'
⍝ ps←⎕NEW Parser '/date= /sw2 /sw3[=]a bc def'  'upper'
⍝ data←ps.Parse 'dsa def /dat=''07/08/28''' ⍝ will return a namespace containing
⍝ data.SW2≡0                                ⍝ 2 strings and set 3 switches
⍝ '07/08/28'≡data.Switch 'DATE'
    
⍝ If a one time parsing a needed the following will do:
⍝ data←(⎕new Parser '/sw1 /sw2...').Parse string
    
    ⎕io←0 ⋄ ⎕ml←1
    LOWER UPPER←'abcdefghijklmnopqrstuvwxyz' ⎕a
    ⎕fx's←upperCase s;b;i' '→0↓⍨1∊b←26>i←LOWER⍳s' '(b/s)←UPPER[b/i]'
    xCut←{⍺←⍵=1↑⍵ ⋄ 1↓¨⍺⊂⍵}     ⍝ exclusive cut
    fixCase←{⍵}                 ⍝ assume none
    sqz←{(b⍲1⌽b←' '=s)/s←' ',⍵} ⍝ remove double spaces, add leading ' '
    if←/⍨
    
    :field ERROR0←700   ⍝ errors are raised starting at this number
    :field DELIMITER    ⍝ the character used to delimit switches
    :field FORCESPACE←1 ⍝ whether a space MUST precede a switch
    
∇ init arg;model;features;b;f
    ⍝ Initialize switch class
 :Access public
 :Implements constructor
 (model features)←2↑⊂⍣(1≡≡arg)+arg
 DELIMITER←⍬⍴model ⍝ 1st char of 'model' delimits switches.
 :If ' '∨.≠features
     Ps←⎕NEW Parser' upper allownospace error=' ⍝ allowed switches for the parser itself
     Pset←Ps.Parse sqz features
     FORCESPACE←~Pset.allownospace
     fixCase←upperCase⍣Pset.upper
     b←0 999≠.<ERROR0←700 Pset.Switch'error'
     'error must be between 0 and 1000'⎕SIGNAL b↓11
     ⎕EX'Ps' ⍝ no need anymore
 :EndIf

    ⍝ Parse the model
 SwTable←↑{(1,1↓<\⍵∊'[=∊')⊂⍵}¨xCut model
 SwTable[;1]←{(-⊥⍨' '=⍵)↓⍵}¨SwTable[;1]
 f←SwTable[;0]←fixCase¨SwTable[;0]~¨' '
 'switches must be unique'⎕SIGNAL 11 if f≢∪f
 'switches must be valid identifiers'⎕SIGNAL 11 if ¯1∊⎕NC⊃f

    ⍝ Create a namespace based on that model
    ⍝ (we could have created a similar class but we would
    ⍝  have had to use ⎕NEW instead of ⎕NS ⎕OR)
 pData←⎕NS''
 ⍎⍕'pData.(',SwTable[;0],')←0'
 pData.Switch←Switch
∇
    
   ⍝ This function is used to return a switch' value, possibly defaulted, e.g.
   ⍝ sw←321 Switch 'abc' ⍝ if 'abc' has not been set 123 is returned.
   ⍝ If /abc=789 was specified in the parsed string the value returned will be ,789, not '789'
    
    ∇ r←{def}Switch s;⎕IO
    ⍝ Return switch' value
      r←1⊃SwD[SwD[;⎕IO←0]⍳⊂,s;]
      :If 0≠⎕NC'def'
          :If 0≡r ⋄ r←def ⍝ use default if not set
          :ElseIf (1≢r)∧∨/1 3 5∊10|⎕DR,def
              r←1⊃⎕VFI r ⍝ make numeric if default is also numeric
          :EndIf
      :EndIf
    ∇
    
    ⍝ Parsing function.
    ⍝ Spaces are used to delimit arguments.
    ⍝ Quotes must be used to include spaces or delimiter in arguments.
    ⍝ Switches can exist with or without [possibly defaulted] value
    ⍝ or be elided. Switches not mentioned are refused.
    ⍝ Shorter names are accepted but '=' MUST be used to supply values.
    
∇ data←Parse arg;Er;s;t;sw;table;parms;swit;swmat;m;val;vnc;set;req;pat;nov;q;∆;i;p;bad;noq
    ⍝ PARSE switches and reset argument. Version 3.01
 :Access public
     ⍝ Account for "
 :If ∨/q←'"'=arg←' ',arg,' '
     t←(1⌽q)⍲s←DELIMITER=arg
     q>←¯1⌽s ⍝ remove those " to ignore
     arg q←t∘/¨arg q ⍝ remove \ where followed by "
     (q/arg)←⊂'''''' ⋄ arg←∊arg
 :EndIf
 Er←ERROR0 ⋄ s←-FORCESPACE
 t←≠\''''=arg ⋄ s←∨\sw←s⌽t<arg⍷⍨(s-1)↑DELIMITER ⍝ QUOTED ⋄ all SWITCHES
 'unbalanced quotes'⎕SIGNAL(¯1↑t)⍴Er
 data←⎕NS ⎕OR'pData' ⋄ table←SwTable[;,0],0
    ⍝ Split the arguments from the switches
 parms←(~s)/arg
 swit←{s←{(-⊥⍨' '=⍵)↓⍵}⍵ ⋄ (⍴s)≠⍴n←fixCase s↑⍨s⍳'=':n((1+⍴n)↓s) ⋄ n 1}¨sw xCut arg
 noq←{∨/b←''''=⍵:1↓(≠\b)/⍵ ⋄ ⍵}
 swmat←' ',⍨↑SwTable[;0]

    ⍝ Process each switch separatly according to their position
 :While 0<⍴swit
     sw val←⊃swit
     m←1⌽'>unknown or ambiguous switch: <',sw
     :If 1≠+/i←((sw,' ')⍷swmat)[;0] ⍝ exact matches
         m ⎕SIGNAL(1≠+/i←(sw⍷swmat)[;0])⍴Er+1
     :EndIf
     vnc←'['=⊃1⊃sw pat←,i⌿SwTable ⍝ '[=]' means 'value not compulsory'
         ⍝ 'pat' is 1 switch group complete with assignment & allowed values if present
     set←'∊'=⊃vnc↓pat ⋄ pat[set/vnc]←'='
         ⍝ We need to check if a value is needed or not allowed
     req←'='=⊃vnc↓pat ⋄ nov←(0∊⍴val)∨t←1≡val
     m←1⌽'>no value allowed for switch <',sw
     m ⎕SIGNAL(req∨t)↓Er+2
     m←1⌽'>value required for switch <',sw
     m ⎕SIGNAL(nov∧req>vnc)⍴Er+3
     table[i⍳1;1]←⊂val←noq val ⍝ remove extra quotes
         ⍝ Valid strings are supplied with the switch: verify them
     :If 0<⍴t←(1+vnc+pat⍳'=')↓pat
         bad←(set∧∧/val∊t)⍱set<∨/(' ',val,' ')⍷' ',t,' '
         m←1⌽'>invalid value for switch <',sw
         m ⎕SIGNAL bad⍴Er+4
     :EndIf
         ⍝ Assign in instance
     ⍎'data.',sw,'←val'
     swit←1↓swit
 :EndWhile

    ⍝ Find each token in the argument and use the delimiter to separate
    ⍝ them. The following allows to deal with '' properly:
 s←parms∊' ' ⋄ ∆←≠\q←parms∊'''' ⋄ i←q∧1⌽q ⋄ t←q>¯1⌽q ⍝ t=1st quote
 p←t∨s>∆ ⍝ DELIMITERS: spaces NOT in quotes OR 1st quote
    ⍝ COMPRESSOR: any double quote OR last quote NOT in ≠\ or double delim.
 parms←(t←(∆<i∨q>1⌽q)⍱p∧1⌽p)/parms ⋄ p←t/p
 data.Arguments←p xCut parms
 data.SwD←table
∇
    
:EndClass