﻿:class CallingTree
⍝
⍝ Lists references calls made by a function. Original code from DanB 1992.
⍝
⍝ This class will allow to produce a list of all the calls made by a function
⍝ as well as cross reference the names in a namespace.
⍝
⍝ Syntax: ⎕new Calls 'namespace[.fn] -FULL -MAXLEVEL -ONLYFNS -TREEVIEW'
⍝     or: instance.Calls 'fn'
⍝
⍝      It will allow to see all objects refered to by a given function
⍝    with repeated procedure for each function called.
⍝
⍝      Each name is preceded by a symbol. Its meaning is
⍝
⍝             G global object (not a visible function)
⍝             F visible function
⍝             R recursive function call
⍝             * function already shown (only without FULL option)
⍝             ○ local  object
⍝             ! unreferenced local object
⍝             l unreferenced label
⍝             L referenced label
⍝             ↑ global local (global localized at an upper level)
⍝
⍝ Example:
⍝        Calls '#.Myfn   -FULL'
⍝
⍝      will show all objects in each fns called by 'Myfn', repeating the
⍝    the procedure for fns already shown (except recursive calls).
⍝    This may seem redondant but consider the following:
⍝
⍝      ∇fn            ∇f1;⎕io        ∇f2
⍝  [1] f1         [1] f2         [1] ⎕io
⍝  [2] f2∇        [2] ∇          [2] ∇
⍝
⍝      f2 refers to ⎕io globally. In 'fn', f2 refers to the global ⎕io in
⍝    the ws. When called by f1 it refers to the local ⎕io in f1. Calls 'fn'
⍝    will produce 2 different outputs for f2, one from 'fn' and one from 'f1'
⍝    because the -FULL switch as been set.
⍝
⍝        Calls 'ns.fn'
⍝
⍝      will show all objects in each fn called by 'fn' but won't show already
⍝    shown functions. In this case the second output for 'f2' won't be present
⍝    and therefore show the local reference to ⎕io.
⍝
⍝        Calls 'ns.fnx -MAX=1'    will only show the fn itself (1st level)
⍝
⍝      Note that ⍎ is being taken in account but that its use may not be
⍝    interpreted properly. Text in functions with ⍎ is scanned between the ⍎
⍝    character and the next ⋄ character on each line where it appears only.
⍝    This also applies to error trapping and some system functions.
⍝
⍝        Calls 'x.fny -ONLYFNS'   will only display the names of the functions
⍝                                 called, not the other objects (e.g. variables).
⍝
⍝        Calls 'n1.fnz -TREEVIEW' will display the result in a treeview form.
⍝
⍝      If no fn name is supplied (only a namespace is given) then the instance is
⍝    prepared and no display occurs immediatly. Subsequent calls to
⍝
⍝    Calls 'fnname -report'       will show a report on the calls made by <fnname>
⍝                                 (by default a tree view object will be used)
⍝
⍝ Other functions:
⍝
⍝    Xref                         will produce a table of names cross-referenced
⍝
⍝ Caveats:
⍝
⍝      The detection of local variables within dfns relies on ← and is wrong for
⍝    multiple assignments (e.g. "a b←1 2")
⍝
⍝      The inclusion of text after ⍎, ⎕NC and ⎕TRAP relies on the presence of
⍝    visible strings and may be wrong if the text is incomplete (e.g. "⍎'a',⍕n)
    
    :include textUtils  ⍝∇:require Utils\textUtils
    
    ALPHABET←'0123456789⎕#.',(0≤⎕NC 256 1⍴⎕AV)/⎕AV
    ⎕io←1 ⋄ CR←⎕av[4] ⋄ ⎕ml←2
    Version←1.22                             ⍝ made into a class & added Xref
    
    :field public  FnRefs
    :field private FnCalls←⍬                 ⍝ none yet
    
    ∇ ctor Xf;Xnl;fn;run;src;b;⍙XREF;class
     ⍝∇ DanB 840921 list all objects of all calls made by argument
     ⍝ -FULL is Full Search (repeat search for already processed fns)
     ⍝ -MAXLEVEL is maximum level to look for
     ⍝ -DETAILS  include variables in reports
     ⍝ -TREEVIEW shows the output in a treeview form
      :Access public
      :Implements constructor
      FnRefs←0 3⍴run←⍙MAXLEVEL ⍙DETAILS ⍙FULL ⍙TREEVIEW ⍙XREF←0
      Xpath←Xf
     ⍝ We can also simply pass a ref as argument
      :If class←9=⎕NC'Xpath'
          Xnl←1↓'-MAXLEVEL= -DETAILS -FULL -TREEVIEW -XREF'∆parse Xf
          ⍙MAXLEVEL←⍙MAXLEVEL ∆default 9999
     
     ⍝ We assume the names to be all in the same [name]space
          Xf←(⎕IO⊃⎕NSI),'.',Xnl
          :If run←9≠⎕NC Xf ⋄ (Xf Xnl)←Xf splitLast'.' ⋄ :EndIf
          Xpath←⍎Xf ⋄
     
    ⍝ If the namespace is a class we turn it into a NS to be able to see its contents
      :OrIf class←9.4∊⎕NC⊂Xf
          src←¯1↓{((∨\b><\b)∧⌽∨\⌽b←'⍝'≠↑¨⍵)/⍵}leftJustify¨⎕SRC Xpath
        ⍝ We must remove all :include, :property and :field statements
          b←':field' ':inclu' ':endpr' ':prope'⍳lowerCase¨6↑¨src
          src←((b∊⍳3)⍱≠\b∊2+⍳2)/src
          Xpath←0 ⎕FIX 1⌽':endnamespace' ':namespace X',src
      :EndIf
      :If ⍙XREF∨class
          fn←Xpath.⎕NL-3 4
        ⍝ For classes we first determine the reference of each function
          FnRefs←fn,(fn∘Xrf∘Xpath.⎕VR¨fn),[1.1]Attributes¨Xpath∘,∘⊂¨fn
          ⎕EX class/'Xpath' ⍝ get rid of temp class
      :EndIf
      ⎕DF'Tree of ',NSPATH←⍕Xf
     
      :If run ⋄ Calls Xnl ⍙DETAILS ⍙FULL ⍙MAXLEVEL ⍙TREEVIEW ⋄ :EndIf
      :If ⍙XREF ⋄ Xref ⋄ :EndIf
    ∇
    
    _MaxLines←5 0 ⍝ comment lines from top/bottom
    
    :property CommentsLines
    ⍝ This property determines how many comment lines from the top and bottom
    ⍝ of each the functions are displayed in Report.
    :access public
        ∇ set n;v
          'must be 1 or 2 positive numbers'⎕SIGNAL 11↓⍨(∧/v≥0)∧(⍴v←,n.NewValue)∊1 2
          _MaxLines←2↑v
        ∇
        ∇ r←get
          r←_MaxLines
        ∇
    :endproperty
    
    ∇ r←Calls Xnl
      :Access public
      :If 82∊⎕DR Xnl
          Xnl←1↓'-DETAILS -FULL -MAXLEVEL= -TREEVIEW'∆parse Xnl ⎕THIS
          ⍙MAXLEVEL ∆default←999
      :Else
          (Xnl ⍙DETAILS ⍙FULL ⍙MAXLEVEL ⍙TREEVIEW)←Xnl
      :EndIf
      :If 0<⍴,Xnl ⍝ there may be a previous result in FnCalls
     ⍝ Initialize tree and output function
          FnCalls←0 4⍴''
          Xtreegen(,⊂Xnl)⍬  ⍝ start the search
      :EndIf
      :If ⍙TREEVIEW ⋄ View ⋄ r←0 0⍴'' ⋄ :Else ⋄ r←Report ⋄ :EndIf
    ∇
    
    ∇ x←Xref;fns;all;o;w;line;i;codes;objs;p;b
⍝ Generate a Cross reference of all names recorded in the instance
      :Access public
      x←1 1⍴'' ⋄ →0↓⍨⍴fns←fns[o←⎕AV⍋⊃fns←FnRefs[;1]]
      w←↑⍴all←all[⎕AV⍋⊃all←(~'⎕⍎'∊⍨↑¨all)/all←∪⊃,/2⊃¨FnRefs[;2]]
      x←⊂w⍴' - - - - :' ⍝ or '----:----|'
      :For i :In ⍳⍴fns
          codes objs←2⊃FnRefs[o[i];]
          b←w≥p←all⍳objs ⋄ line←w⍴' . . . . :' ⋄ line[b/p]←b/codes
          x,←⊂line
      :EndFor
⍝ Prepare output
      o←⍉⊃all,¨w↓¨(⌈⌿⊃w←⍴¨all)⍴¨(⍴all)⍴' .'
      x←⍕2 2⍴''o(⊃' ',fns)(⊃x)
    ∇
    
    
⍝    --- Utilities
    
     ⍝ Partionned ∨\, ≠\  fns
    ∇ r←s OS p;t ⍝ ∨\
      s←s∨p ⋄ r←≠\s\t≠¯1↓0,t←s/p
    ∇
    ∇ r←s US p;t ⍝ ≠\
      r←≠\p≠s\t≠¯1↓0,t←s/≠\¯1↓0,p
    ∇
    
    ∇ r←del toMat str
     ⍝ Use delimiter to split string. Multiple dels=1
      r←⊃(⊂'')~⍨str splitOn del
    ∇
    
    ∇ (type n)←fns Xrf fn;dfn;loc;o;t;l;B;p;⎕IO;lab;T;C;D;E;nl;c;Alf;bl;b;lp;rp;sp;dcode;cut;type;ec
     ⍝ Find a fn's local and global references given its ⎕VR.
     ⍝ Modified 2005 for Dyalog's (header syntax) and 2007 for classes.
      :Access public shared
      ⎕IO←1 ⋄ Alf←ALPHABET ⋄ type←lab←loc←,n←0⍴bl←' ' ⋄ nl←0 ⋄ cut←{(⍵ splitOn ⍺)~⊂''}
      →(82∊⎕DR fn)↓0 ⍝ skip special fns like 'dsa'∘,
      t←1∊⍴⍴fn ⋄ fn←1↓,CR,fn ⍝ ⎕CR also allowed
     
     ⍝ Is this fn a Dfn?
      dfn←'}'∊¯1↑(-p⊥p←fn∊'∇ ',CR)↓fn ⋄ Alf←Alf,dfn/'⍺⍵←'
      :If t∨dfn ⋄ fn←(fn⍳'∇{'[t⌈2×dfn])↓fn ⋄ :EndIf
     
      :If ~dfn
     ⍝ Not a Dfn, find local references
     ⍝ Because DYW now (2004) has a multiple header syntax we must account for it,
     ⍝ e.g. {(a b)}←(d e) foo (f g)   -or-   (f1 op f2)arg
          t←(n←¯1+⌊/l⍳'⍝;')↑l←t\(t←~l∊'{}')/l←(nl←¯1+fn⍳CR)↑fn
          loc←'; 'cut{(¯1+⍵⍳'⍝')↑⍵}n↓l   ⍝ locals
          loc,←bl cut((¯1+p←-n⊥n←'←'≠t)↓t)~'()'      ⍝ result(s)
     ⍝ The header is trickier as we must remove the fnname. Here a the various possibilities:
     ⍝ [(left args)] (operator args) (right args) where the parens may be omitted with only 1 name
     ⍝ We use the parens and spaces as delimiters to find the middle section:
          sp←bl=l←Xqz p↑t ⋄ lp←l∊'(' ⋄ rp←l∊')'
          b←≠\(2≠+/b),b←1↓lp∨((sp∨¯1⌽rp)>1⌽lp)>≠\lp∨rp
          :If 1↑b/lp ⋄ b←b\≠\b/sp ⋄ :EndIf ⍝ work the op's name using the spaces
          l←(~b)/l ⋄ b←~l∊'()'
          loc←∪loc,bl cut b\b/l
      :EndIf
      nl←CR=n←nl↓fn
     ⍝ Text, Body, Comments, statement Delimiters & ⍎
      T←nl US n∊'''' ⋄ E←n='⍎'        ⍝ Text & ⍎
      B←C⍱T>C←nl OS T<('⍝∇∪'⍷n)<n∊'⍝' ⍝ allow to denote ⍝∇∪ var
     
     ⍝ V11 introduces some special :Statements that we take into account:
      B←T⍱C←C∨nl OS B∧↑∨/⍷∘n¨':Signature' ':Implements' ':Access'
     
     ⍝ Mask out dfns inside a fn
      B←B>dcode←×(+\B∧n∊'{')-+\B∧n∊'}' ⋄ D←nl∨B∧n∊'⋄' ⋄ l←0
      :If ~dfn
     ⍝ Account for control structures
          T←':If' ':AndIf' ':OrIf' ':ElseIf' ':EndIf' ':ForEach' ':InEach' ':EndFor'
          T←T,':Select' ':CaseList' ':EndSelect' ':While' ':EndWhile'
          T←T,':Repeat' ':EndRepeat' ':Until' ':Trap' ':EndTrap'
          T←T,':GoTo' ':Return' ':Continue' ':Leave'
          :For c :In T
              l←l∨t∨(-⍴c)⌽t←c⍷n  ⍝ find start/end of each :keyword
          :EndFor
          T←':Case' ':Else' ':End' ':For' ':In' ⍝ special case for other shorter names
          :For c :In T
              l←l∨t∨(-⍴c)⌽t←l<c⍷n
          :EndFor
          B←B>≠\l ⍝ mask out from Body
          l←B∧n∊':'
      :EndIf
      E←B∧E ⋄ ec←E∨(¯12⌽n⍷⍨'⎕TRAP←')∨(n⍷⍨'⎕NC ')∨n⍷⍨'⎕NC''' ⋄ E←∨/E
      B←B∨C<D OS ec ⍝ executable code
      c←t←p←T←C←D←1 ⍝ recoup space
      :If ~dfn ⍝ find Labels
          p←(D←⌽(⌽nl)OS⌽l)/n ⋄ l←dcode<D\t⍱≠\t←p∊'[]' ⍝ remove line [#]
          lab←': 'cut l/n ⋄ B←B>l ⍝ exclude labels from Body
      :EndIf
     ⍝ Find global references
      t←(B⍲n∊Alf)∨'.⎕'⍷n      ⍝ find delimiters
      n[t/⍳⍴t]←bl             ⍝ replace by blank
      n←Xqz(('. '⍷n)⍱¯1⌽' .'⍷n)/n ⍝ remove 1st & last dot and squeeze text
      n←(¯1⌽~≠\(1+t)/t←'←'=n)\n←(~' ←'⍷n)/n
      t←'←'∊¨n←(~(↑¨n)∊⎕D)/n←∪bl cut n   ⍝ remove any space before ←
      :If dfn ⋄ loc←(¯1↓¨t/n),(o∊n)/o←,¨'⍺⍵' ⋄ :EndIf
      n←∪(-t)↓¨n              ⍝ remove ← (could gather info for line # here)
      o←⊃,/l←loc lab fns
      type←(∊(⍴¨l)⍴¨'○LF'),'G'
      type←type[o⍳n]
      t←~loc∊n ⋄ n type,←t∘/¨loc'!'
      t←~lab∊n ⋄ n,←t/lab ⋄ type,←t/'l'
      →E↓0
      n type←n type,¨'⍎ '
    ∇
    
    ∇ refs←{st}Xgensymb br;i;fn;cr
     ⍝ Generate new symbols of last function in br
      :If (i←FnRefs[;1]⍳fn←¯1↑br)>1↑⍴FnRefs
          cr←Xpath.⎕CR↑fn
          FnRefs⍪←fn,((Xpath.⎕NL-3 4)Xrf cr)(cr Attributes Xpath,fn)
      :EndIf
      refs←1⊃FnRefs[i;2]
      refs←Xcompare st refs br ⍝ correct references
     ⍝ Gather results
      FnCalls⍪←fn,(⍴br),refs(3⊃,FnRefs[i;]) ⍝ update Globals
    ∇
    
    ∇ refs←Xcompare(stack refs branch);globals;codes;gnam;c;o;n;t;fn;nam;allf
     ⍝ Compare list and adjust global definitions
     ⍝ Any global or fn found localized up the stack will be marked as '↑'
      →0↓⍨+/globals←'GF'∊⍨1⊃refs ⋄ gnam←globals/2⊃refs
      c o←⌽¨⊃,¨/stack,⊂⍬ ⍬
      :For nam :In gnam
          :If '○'∊1↑(n←+/∧\t∊'↑G')↓t←(o∊⊂nam)/c
              ((1,fn⍳⍨2⊃refs)⊃refs)←'↑'
          :EndIf
      :EndFor
    ⍝ Change F into R for all recursive calls
      :If 0<⍴fn←('F'=codes←1⊃refs)/allf←2⊃refs
    ⍝ We know these fns are NOT localized (cause we'd have found up there)
        ⍝ If we do not want a FULL search we replace F by * to denote 'Done, see above'
          :If ⍙FULL<0<⍴nam←(fn∊FnCalls[;1])/fn ⋄ codes[allf⍳nam]←'*' ⋄ :EndIf
        ⍝ For recursive fns all we need to do is look up the branch to see if it's there
          :If 0<⍴nam←(fn∊branch)/fn ⋄ codes[allf⍳nam]←'R' ⋄ :EndIf
          (1⊃refs)←codes
      :EndIf
    ∇
    
    ∇ Xtreegen(Xbr Xst);symbs;i;fns
     ⍝ Generate symbols of the last fn in the BRanch and those of its sub-routines
     ⍝ semi-globals: Xnl,  name list of all fns processed so far
     ⍝
      symbs←Xst Xgensymb Xbr
      :If ⍙MAXLEVEL>⍴Xbr ⍝    max level reached ?
          fns←('F'=1⊃symbs)/2⊃symbs
          :For i :In ⍳⍴fns ⋄ Xtreegen(Xbr,fns[i])(Xst,⊂symbs) ⋄ :EndFor
      :EndIf
    ∇
    
    ∇ at←{cr}Attributes(path fn);c;b
    ⍝ Return the attributes of a fn given its ⎕CR
    ⍝ Those are the top and bottom comments, its class attributes
    ⍝ and the size and header syntax (3∊)
      :If 0∊⎕NC'cr' ⋄ cr←path.⎕CR fn ⋄ :EndIf
      at←0 ''((path.⎕SIZE fn),1⊃,path.⎕AT fn)
      (1⊃at)←2⍴⊂cr←1 0↓⎕FMT leftJustify cr ⍝ ⎕fmt because cr may be vector (eg {⍵})
      :If 0<↑⍴cr
          (1⊃at)←trimEnds¨((∧\c)⌿cr)((b←⌽∧\⌽c←'⍝'=cr[;1])⌿cr)
          →0↓⍨4<1↓⍴c←lowerCase(':'=cr[;1])⌿cr
          c←' 'splitOn⍨,(':impl' ':acce'∊⍨↓c[;⍳5])⌿c
          (2⊃at)←('public' 'shared' 'constructor' 'destructor'∊c)/'PSCD'
      :EndIf
    ∇
    
    ∇ rep←Report;fn;lev;refs;atts;tb;grab;i;from;⎕IO
    ⍝ Character based listing
      rep←⍴⎕IO←0 ⋄ tb←1 ¯1×_MaxLines ⋄ grab←{n←(×⍺)×(|⍺)⌊1↑⍴⍵ ⋄ ↓(n,1↓⍴⍵)↑⍵}
      :For i :In ⍳1↑⍴FnCalls
          fn lev refs atts←FnCalls[i;]
          from←'→',⍨(i-(⌽0,i↑FnCalls[;1])⍳lev-1)⊃' ',FnCalls[;0]
          rep,←CR,CR,('Level ',⍕lev),': ',from,fn,,CR,⊃⊃,/tb grab¨↑atts
          rep,←,CR,Xft refs
      :EndFor
      rep←2↓rep
    ∇
    
    ∇ View;tree;form;font;rc;⎕WX;details
     ⍝ Display tree in a treeview window
      :Access public
      ⎕WX←1 ⍝ so we can see inside
      tree←FnCalls[;⍳2] ⍝∇∪ form ⍝ we use this name locally
     ⍝ Produce details?
      :If ⍙DETAILS
          details←↓¨(99-3×FnCalls[;2])Xft¨FnCalls[;3]
          tree⍪←⊃⍪/(tree[;2]+1),[1.1]⍨¨details
          tree←tree[⍋⍒∊(1+∊⍴¨details)↑¨1;]
      :EndIf
      font←'Dyalog Std' 16 1 0 0 400 0 0
      :If 2=⎕SE.⎕NC'FontObj' ⋄ font←⎕SE.FontObj ⋄ :EndIf
      rc←(999⌊99⌈18×1↑⍴tree),99⌈999⌊13×⌈⌿(∊⍴¨tree[;1])+2×tree[;2]
      'form'⎕WC'form' 'Tree View of function calls'(20 20)rc('font'font)('coord' 'pixel')
      'form.exit'⎕WC'button' ''(0 0)(1 1)('cancel' 1)('event' 'select' 1)
      'form.tv'⎕WC'treeview'(tree[;1])(0 0)(100 100)('depth'(tree[;2]-1))('coord' 'prop')
      2 ⎕NQ¨'form.tv' 302∘,¨⍳1↑⍴tree
      ⎕DQ'form'
    ∇
    
    ∇ mat←{width}Xft(symb names);p;w;f;d
     ⍝ Reformat mat
      :If 2≠⎕NC'width' ⋄ width←⎕PW ⋄ :EndIf
      symb names←(⍙DETAILS∨symb∊'RF*')∘/¨symb names
      mat←⊃symb,¨names
      mat←mat[⎕AV⍋0 1↓mat;]
      w←5×⌈(1↓d←⍴mat←' ',' ',¯1⌽':',1⌽mat)÷5 ⍝ list's dims
      mat←((f×p←⌈d[⍳1]÷f←⌊width÷w),w)↑mat    ⍝ adjust dims
      mat←(p,f×w)⍴mat                        ⍝ optimized formatting
      mat←(0,-⊥⍨' '∧.=mat)↓mat               ⍝ remove extra columns
    ∇
    
    ∇ r←Xqz s
     ⍝ danb 831012 squeeze out extra spaces
      r←1↓(r⍲1⌽r←s∊' ')/s←,' ',s
    ∇
    
:EndClass